require 'faraday'
require 'json'

module Mycrm
  # this module adds connectivity methods to the descendant's class
  module Connectable
    TOKEN_KEY = 'mycrm_auth_token_key'
    DEFAULT_TOKEN_EXPIRE_IN_STORAGE = 600 # 10 minutes

    def self.extended(descendant)
      descendant.send(:extend, ExtendedMethods)
    end

    class << self
      #create the connection with basic configurations
      def connection
        Faraday.new(url: Mycrm.configuration.base_uri) do |faraday|
          faraday.request :url_encoded
          if Mycrm.configuration.logger
            faraday.response :logger, Mycrm.configuration.logger, bodies: true
          end
          yield(faraday) if block_given?
          faraday.adapter Faraday.default_adapter
        end
      rescue Errno::ECONNREFUSED => e
        raise ConnectionError, e
      end

      def token
        @token || fetch_token
      end

      def fetch_token
        storage = Mycrm.configuration.storage # Redis or any strategy that implement set and get

        if storage
          token = storage.get(TOKEN_KEY)
          if token && !(token.nil? || token.empty?)
            token
          else
            token_storage_expiry = Mycrm.configuration.token_storage_expiry || DEFAULT_TOKEN_EXPIRE_IN_STORAGE
            token = fetch_token_from_mycrm
            storage.set(TOKEN_KEY, token, ex: token_storage_expiry)
            token
          end
        else
          fetch_token_from_mycrm
        end
      end

      def fetch_token_from_mycrm
        connection.post('/Login', credentials).body
      end

      def credentials
        %w(username password).each_with_object({}) do |f, o|
          o[f] = Mycrm.configuration.send(f)
        end
      end

      def session
        @token = token
        yield if block_given?
      ensure
        @token = nil
      end
    end

    # private module to be extended
    module ExtendedMethods

      def connection
        Mycrm::Connectable.connection do |faraday|
          faraday.headers['Authorization'] = "Bearer #{skim(Mycrm::Connectable.token)}"
          api_key = Mycrm.configuration.api_key
          faraday.headers['X-MyCrm-ApiKey'] = api_key if api_key
        end
      end

      def respond(response)
        raise ResponseError, skim(response.body) unless response.success?
        raise ResponseError, 'Not found' if 'null'.eql?(skim(response.body))
        parse(response.body)
      end

      def parse(body)
        JSON.parse(body)
      rescue JSON::ParserError
        raise ResponseError, "#{skim(body)} is not in a valid format"
      end

      def skim(string)
        string.to_s.gsub(/\A"|"\Z/, '')
      end

      %w(get post put delete).each do |method|
        define_method method do |uri, body = {}, query = {}|
          response = connection.send(method, uri, query) do |req|
            req.headers['Content-Type'] = 'application/json'
            req.body = JSON.generate(body)
            req.options.timeout = 120
            req.options.open_timeout = 20
          end
          respond(response)
        end
      end
    end
  end
end
