module Mycrm
  # facades
  module Facade
    def self.included(descendant)
      descendant.send(:include, InstanceMethods)
      descendant.send(:extend, ClassMethods)
    end

    module InstanceMethods
      def self.included(descendant)
        descendant.class_eval do
          attr_accessor :delegates
        end
      end

      def initialize(attributes = {})
        self.delegates = {}
        self.class.delegate_classes.each do |name, delegate|
          delegates[name] = delegate[:class].new attributes[name]
        end
      end

      def to_json
        delegates.each_with_object({}) { |(name, instance), r| r[name] = instance.to_json }
      end

      def run
        {}.tap do |response|
          Mycrm::Connectable.session do
            delegates.each do |name, instance|
              refresh_relation(name)
              response[name] = yield(instance)
            end
          end
        end
      end

      def refresh_relation(name)
        self.class.relations(name).to_a.each do |pair|
          delegates[name].send("#{pair.first}=", relative_value(*pair.last.split('.'))) rescue nil
        end
      end

      def relative_value(delegate_name, field)
        delegates[delegate_name.to_sym].send(field)
      end

      def method_missing(name, *args, &block)
        return delegates[name] if delegates.key?(name)
        delegate = delegates.values.detect { |d| d.respond_to?(name) }
        delegate ? delegate.send(name, *args) : super
      end

      def respond_to_missing?(name, include_private = false)
        delegates[name] || delegates.detect { |d| d.respond_to?(name) } || super
      end
    end

    module ClassMethods
      def self.extended(descendant)
        descendant.singleton_class.class_eval do
          attr_accessor :delegate_classes
        end
      end

      def delegate(name, clazz, relations = {})
        self.delegate_classes ||= {}
        delegate_classes[name] = { class: clazz, relations: relations }
      end

      def relations(delegate)
        delegate_classes[delegate][:relations]
      end

      [:find, :list].each do |method|
        define_method method do |*args|
          run method, *args
        end
      end

      def run(method = :find, *args)
        attributes = {}.tap do |response|
          Mycrm::Connectable.session do
            self.delegate_classes.each do |name, delegate|
              res = if block_given?
                yield(name, delegate, response)
              elsif delegate[:class].respond_to? :find
                delegate[:class].send(method, *args)
              end
              response[name] = res.is_a?(Array) ? res.map(&:to_json) : res.to_json
            end
          end
        end
        new(attributes)
      end
    end
  end
end
