require 'virtus'
require 'virtus_convert'
require_relative 'connectable'

module Mycrm
  # models
  module Model
    def self.included(descendant)
      descendant.send(:include, Virtus.model)
      descendant.send(:include, InstanceMethods)
      descendant.send(:extend, Connectable)
      descendant.send(:extend, ClassMethods)

      descendant.class_eval do
        attr_accessor :error
      end
    end

    # provides methods for the instance
    module InstanceMethods
      def initialize(attributes = nil)
        partitions = (attributes || []).partition{ |field, _| attribute_set[field].nil? }
        super(partitions.last.to_h)
        partitions.first.each { |field, value| send("#{field}=", value) rescue nil }
      end

      def create!
        mass_assign(self.class.execute(:post, nil, to_json))
      end

      def update!
        mass_assign(self.class.execute(:put, id, to_json))
      end

      def delete!
        self.class.execute(:delete, id, nil)
      end

      def to_json
        VirtusConvert.new(self).to_hash.delete_if(&method(:swoop))
      end

      def swoop(_, v)
        (v.is_a?(Hash) ? v.delete_if(&method(:swoop)) : v.to_s).empty?
      end

      def mass_assign(json)
        attribute_set.set(self, json)
      end

      # safe methods for the principal crud methods
      %w(create update delete).each do |method|
        define_method method do
          begin
            send("#{method}!")
            true
          rescue => e
            self.error = e.message
            false
          end
        end
      end
    end

    # provides method at class level
    module ClassMethods
      def self.extended(descendant)
        descendant.singleton_class.class_eval do
          attr_accessor :class_endpoint, :alowed_methods, :attribute_to
        end
      end

      def list(id = nil)
        to_list(execute(:get, id))
      end

      def list_by(key, value, use_key_as_path = false)
        path = use_key_as_path ? key : nil
        to_list(execute(:get, path, {}, key => value))
      end

      def find(id)
        new(execute(:get, id))
      end

      def find_by(key, value, use_key_as_path = false)
        path = use_key_as_path ? key : nil
        new(execute(:get, path, {}, key => value))
      end

      def execute(action, path, body = {}, query = {})
        raise ApiError, 'Method not allowed' if !alowed_methods.nil? && !alowed_methods.include?(action)
        Mycrm.log(:info, %{
--------- #{self.name} ---------
method:   #{action}
path:     #{uri(path, body)}
body:     #{body}
query:    #{query}
-------------------------------})
        send(action, uri(path, body), body, query).tap { |response|  Mycrm.log(:info, "response: #{response}") }
      end

      def uri(relative = nil, parameters = {})
        format([class_endpoint, relative.to_s].compact.join('/').chomp('/'), parameters)
      rescue KeyError => e
        raise ApiError, e
      end

      def to_list(res)
        raise ApiError, 'The output is not compatible.' unless res.is_a?(Array)
        res.each_with_object([]) { |j, list| list << new(j) }
      end

      # macro method to setup the endpoint that will be used to connect the API
      def endpoint(endpoint)
        self.class_endpoint = endpoint
      end

      # macro method to limit the method type
      def only(*methods)
        self.alowed_methods = methods.collect(&:to_sym)
      end

      def to(mode)
        raise ApiError, "The conversion type '#{mode}' is not compatible." unless [:underscore, :camelized].include?(mode)
        self.attribute_to = mode
      end

      # macro method to create a new attribute of type Object
      # and relative accessors
      def attribute(name, klazz, options = {})
        if options[:embedded]
          options.merge!(default: {})
          define_embedded_methods(name, klazz)
        end

        define_formatted_methods name

        super

        alias_method(options[:as], name) if options[:as]
      end

      # Defines accessors for embetted attributes
      #
      # class Phone
      #   attribute :home
      #   attribute :work
      #   attribute :mobile
      # end
      #
      # class Contact
      #  attribute :phone, embedded:true
      # end
      #
      # contact = Contact.new
      # contact.phone_home = 'xxx'
      # contact.home_phone = 'xxx'
      #
      # private
      def define_embedded_methods(field, klazz)
        klazz.allowed_writer_methods.each do |att|
          ['%{field}_%{attribute}', '%{attribute}_%{field}'].each do |template|
            define_embedded_method template, field, att.to_s.chomp('=')
          end
        end
      end

      def define_embedded_method(template, field, attribute)
        accessor = format(template, field: field.to_s, attribute: attribute)

        define_method "#{accessor}=" do |val|
          instance_variable_get("@#{field}").send("#{attribute}=", val)
        end

        define_method accessor do
          instance_variable_get("@#{field}").send(attribute)
        end
      end

      def define_formatted_methods(field)
        return unless attribute_to
        alias_name = field.send("to_#{attribute_to}")

        define_method "#{alias_name}=" do |val|
          send("#{field}=", val)
        end

        define_method alias_name do
          send(field)
        end
      end

      def method_missing(name, *args, &block)
        name =~ /^(find|list)_by_(.*)$/ ? send("#{Regexp.last_match[1]}_by", Regexp.last_match[2], *args) : super
      end

      def respond_to_missing?(name, include_private = false)
        name !~ /^(find|list)_by_(.*)$/ ? super : true
      end
    end
  end
end
